#ifndef ARMARX_ROBOTSTATE_ROBOTSTATE_SLICE
#define ARMARX_ROBOTSTATE_ROBOTSTATE_SLICE

#include <RobotAPI/interface/units/KinematicUnitInterface.ice>
#include <RobotAPI/interface/units/PlatformUnitInterface.ice>
#include <RobotAPI/interface/robotstate/PoseBase.ice>

module armarx
{
    /**
     * Available VirtualRobot joint types.
     */
    enum JointType
    {
        ePrismatic,
        eFixed,
        eRevolute
    };

    /**
     * Distributed reference counting interface for the SharedRobotInterace
     * instances.
     */
    interface SharedObjectInterface
    {
        void ref();
        void unref();
    };

    sequence<string> NameList;

    /**
     * The data stored in a VirtualRobot::RobotNodeSet but accessable over Ice.
     */
    class RobotNodeSetInfo
    {
        NameList names;
        string name;
        string tcpName;
        string rootName;
    };

    /**
     * The SharedRobotNodeInterface provides access to a limited amount of
     * VirtualRobot::RobotNode methods over the Ice network.
     */
    interface SharedRobotNodeInterface extends SharedObjectInterface
    {

        ["cpp:const"] idempotent
        float getJointValue();
        ["cpp:const"] idempotent
        string getName();

		["cpp:const"] idempotent
        PoseBase getLocalTransformation();
        
        ["cpp:const"] idempotent
        PoseBase getGlobalPose();
        ["cpp:const"] idempotent
        PoseBase getGlobalPoseJoint();

        ["cpp:const"] idempotent
        JointType getType();
        ["cpp:const"] idempotent
        Vector3Base getJointTranslationDirection();
        ["cpp:const"] idempotent
        Vector3Base getJointRotationAxis();

        ["cpp:const"] idempotent
        bool hasChild(string name, bool recursive);
        ["cpp:const"] idempotent
        string getParent();
        ["cpp:const"] idempotent
        NameList getChildren();
        ["cpp:const"] idempotent
        NameList getAllParents(string name);

        ["cpp:const"] idempotent
        float getJointValueOffest();
        ["cpp:const"] idempotent
        float getJointLimitHigh();
        ["cpp:const"] idempotent
        float getJointLimitLow();

    };


    /**
     * The SharedRobotInterface provides access to a limited amount of
     * VirtualRobot::Robot methods over the Ice network.
     */
    interface SharedRobotInterface extends SharedObjectInterface
    {
        SharedRobotNodeInterface* getRobotNode(string name);
        SharedRobotNodeInterface* getRootNode();
        bool hasRobotNode(string name);
        NameList getRobotNodes();

        RobotNodeSetInfo getRobotNodeSet(string name);
        NameList getRobotNodeSets();
        bool hasRobotNodeSet(string name);

        string getName();
        string getType();

        PoseBase getGlobalPose();
        NameValueMap getConfig();
    };


    /**
     * The interface used by the RobotStateComponent which allows creating
     * snapshots of a robot configuration or retrieving a proxy to a constantly
     * updating synchronized robot.
     */
    interface RobotStateComponentInterface extends KinematicUnitListener, PlatformUnitListener
    {
        /**
         * @return proxy to the shared robot which constantly updates all joint values
         */
        SharedRobotInterface* getSynchronizedRobot();
        /**
         * @return proxy to a copy of the shared robot with non updating joint values
         */
        SharedRobotInterface* getRobotSnapshot(string time);
    };
};

#endif
