#ifndef MOTIONCONTROL_H
#define MOTIONCONTROL_H

#include <Core/statechart/Statechart.h>
#include <RobotAPI/core/RobotStatechartContext.h>

namespace armarx
{

namespace MotionControl
{
/**
    \defgroup MotionControl ArmarX MotionControl
    \section Overview Overview of the implemented functionality in MotionControl
    All functionality is implemented with State to provide a convenient interface
    for the non-instant move functions (States receive events when desired targets
    are reached).<br/>
    The following functionality is available in the MotionControl:
    - Direct kinematics: move joints to a desired configuration (armarx::MotionControl::MoveJoints)
    - Velocity control: set the desired velocities for joints (armarx::MotionControl::MoveJointsVelocityControl)
    - Inverse kinematics: move the TCP of a kinematic chain to a desired pose (armarx::MotionControl::MoveTCPPoseIK)
    - Trajectory: move the TCP of a kinematic chain along a trajectory defined by a list of poses (armarx::MotionControl::MoveTCPTrajectory)
    - Stop: stop all motion of the robot (armarx::MotionControl::StopRobot)

    \section HowTo How to use the states
    To use one of the MotionControl states, you have to make it a substate of your own state (see ArmarXCore::Statechart). To set its parameters,
    map them from you own state to it. The standard configuration is set in file ${ArmarX_Home}/data/HumanoidRobotAPIConfigs/stateconfigs/MotionControl.xml.

    Two kinds of kinematic entities are controlled using the MotionControl: single joints and kinematic chains. A kinematic chain is a list of joints
    that usually belong to a part of the robot, e.g. an arm. The joints and the kinematic chains are all defined in the xml files that describe the
    kinematic model of the robot (e.g. ${ArmarX_Home/data/ArmarIV/RobotModel/ArmarIV.xml and the files for the subparts). Look in these files to find out the name of
    the joint or kinematic chain you want to move.

    \subsection UsingMoveJoints Using armarx::MotionControl::MoveJoints
    This state moves the specified joints to the desired configuration.
    Input parameters:
    - jointNames: the names of the joints to be moved
    - targetJointValues: the desired joint values
    - targetTolerance: tolerance defining how precisely the joint position has to be reached
    - timeoutInMs: a timeout after which the attempt to move is aborted
    Output parameters:
    - jointDistancesToTargets: the distance of each joint to the desired configuration

    \subsection UsingMoveJointsVelocityControl Using armarx::MotionControl::MoveJointsVelocityControl
    This state moves the specified joints with the desired velocity.
    Input parameters:
    - jointNames: the names of the joints to be moved
    - targetJointVelocities: the desired joint velocities
    - targetJointVelocityTolerance: tolerance defining how precisely the joint velocity has to be reached
    - timeoutInMs: a timeout after which the attempt is aborted
    Output parameters:
    - jointVelocitiesDistancesToTargets: the difference between the desired and actual joint velocities

    \subsection UsingMoveTCPPoseIK Using armarx::MotionControl::MoveTCPPoseIK
    Moves the TCP of the specified kinematic chain to the desired pose if possible.
    Input parameters:
    - kinematicChainName: the name of the kinematic chain that is used
    - targetTCPPosition: the target position for the TCP
    - targetTCPOrientation: the target orientation
    - targetPositionDistanceTolerance: tolerance for the position to decide if the motion was successfull
    - targetOrientationDistanceTolerance: tolerance for the orientation to decide if the motion was successfull
    - ikWithOrientation: Consider orientation for IK computation. if false, the rotation of the TCP is not considered.
    - timeoutInMs: a timeout after which the motion is aborted
    Output parameters:
    - TCPDistanceToTarget: kartesian distance between TCP and target position
    - TCPOrientationDistanceToTarget: the difference between desired and actual orientation of the TCP

    \subsection UsingMoveTCPTrajectory Using armarx::MotionControl::MoveTCPTrajectory
    Moves the TCP of the specified kinematic chain along the defined poses.
    Input parameters:
    - kinematicChainName: the name of the kinematic chain that is used
    - targetTCPPositions: the list of target positions for the TCP
    - targetTCPOrientations: the list of target orientations
    - targetPositionDistanceTolerance: tolerance for the position to decide if the motion was successfull
    - targetOrientationDistanceTolerance: tolerance for the orientation to decide if the motion was successfull
    - ikWithOrientation: Consider orientation for IK computation. if false, the rotation of the TCP is not considered.
    - timeoutInMs: a timeout after which the motion is aborted
    Output parameters:
    - TCPDistanceToTarget: kartesian distance between TCP and target position
    - TCPOrientationDistanceToTarget: the difference between desired and actual orientation of the TCP

    \subsection UsingStopRobot Using armarx::MotionControl::StopRobot
    Stops all motion of the robot. No parameters.

*/

    /**
    *  \class MotionControlHandler
    *  \ingroup MotionControl
    *  MotionControlHandler is the remote state interface for the MotionControl-
    *  States. It contains convenience states that allow to move parts of the robot,
    *  i.e. direct and inverse kinematics for arbitrary kinematic chains and predefined
    *  parts like the arms.
    */
    class MotionControlOfferer : public RemoteStateOfferer<RobotStatechartContext>
    {
    public:
        MotionControlOfferer();
        void onInitRemoteStateOfferer();
        void onConnectRemoteStateOfferer();
        std::string getStateOffererName() const { return "MotionControl"; }
    };


    DEFINEEVENT(EvJointTargetReached)
    DEFINEEVENT(EvJointVelocitiesTargetReached)
    DEFINEEVENT(EvCalculationDone)
    DEFINEEVENT(EvTimeout)


    // test states

    struct MotionControlTestState : StateTemplate<MotionControlTestState>
    {
        void defineParameters();
        void defineSubstates();
    };


    struct MotionControlTestStateIK : StateTemplate<MotionControlTestStateIK>
    {
        void defineParameters();
        void defineSubstates();
    };



    /**
    *  \ingroup MotionControl
    *   Move the joints of a kinematic chain to the desired values.
    *   \param jointNames the names of the joints to be moved
    *   \param targetJointValues the desired joint values
    *   \param targetTolerance tolerance defining how precisely the joint position has to be reached
    *   \param timeoutInMs a timeout after which the attempt to move is aborted
    */
    struct MoveJoints : StateTemplate<MoveJoints>
    {
        RobotStatechartContext* context;
        ConditionIdentifier targetReachedCondition;
        ActionEventIdentifier timeoutEvent;
        void defineParameters();
        void onEnter();
        void onBreak();
        void onExit();
    };


    /**
    *  \ingroup MotionControl
    *   Set the velocities of the joints of a kinematic chain.
    *   \param jointNames the names of the joints to be moved
    *   \param targetJointVelocities the desired joint velocities
    *   \param targetJointVelocityTolerance tolerance defining how precisely the joint velocity has to be reached
    *   \param timeoutInMs a timeout after which the attempt is aborted
    */
    struct MoveJointsVelocityControl : StateTemplate<MoveJointsVelocityControl>
    {
        RobotStatechartContext* context;
        ConditionIdentifier targetVelocitiesReachedCondition;
        ActionEventIdentifier timeoutEvent;
        void defineParameters();
        void onEnter();
        void onExit();
    };


    /**
    *  \ingroup MotionControl
    *   Move the TCP to a desired pose.
    *   \param kinematicChainName the name of the kinematic chain that is used
    *   \param targetTCPPosition the target position for the TCP
    *   \param targetTCPOrientation the target orientation
    *   \param targetPositionDistanceTolerance tolerance for the position to decide if the motion was successfull
    *   \param ikWithOrientation: Consider orientation for IK computation. if false, the rotation of the TCP is not considered.
    *   \param targetOrientationDistanceTolerance tolerance for the orientation to decide if the motion was successfull
    *   \param timeoutInMs a timeout after which the motion is aborted
    */
    struct MoveTCPPoseIK : StateTemplate<MoveTCPPoseIK>
    {
        RobotStatechartContext* context;
        ConditionIdentifier targetReachedCondition;
        ActionEventIdentifier timeoutEvent;
        void defineParameters();
        void defineSubstates();
        void onEnter();
        void onExit();
    };


    DEFINEEVENT(EvLastPointReached)
    DEFINEEVENT(EvLastPointNotYetReached)

    /**
    *  \ingroup MotionControl
    *   Move the TCP along a trajectory of poses.
    *   \param kinematicChainName the name of the kinematic chain that is used
    *   \param targetTCPPositions the list of target positions for the TCP
    *   \param targetTCPOrientations the list of target orientations
    *   \param targetPositionDistanceTolerance tolerance for the position to decide if the motion was successfull
    *   \param targetOrientationDistanceTolerance tolerance for the orientation to decide if the motion was successfull
    *   \param ikWithOrientation: Consider orientation for IK computation. if false, the rotation of the TCP is not considered.
    *   \param timeoutInMs a timeout after which the motion is aborted
    */
    struct MoveTCPTrajectory : StateTemplate<MoveTCPTrajectory>
    {
        RobotStatechartContext* context;
        ConditionIdentifier lastTargetReachedCondition;
        ActionEventIdentifier timeoutEvent;
        ActionEventIdentifier trajectoryPointCounterID;
        void defineParameters();
        void defineSubstates();
        void onEnter();
        void onExit();
    };




    /**
    *  \ingroup MotionControl
    * Stop all motion of the robot. It sets all velocities to 0. No Smooth stopping
    * is implemented here.
    */
    struct StopRobot : StateTemplate<StopRobot>
    {
        RobotStatechartContext* context;

        void onEnter();
    };


    /**
    *  \ingroup MotionControl
    * Closes the given hand of the robot
    * \param useLeftHand True if left hand should be closed, false for the right hand
    */
    struct DoPreshape : StateTemplate<DoPreshape>
    {
        void defineParameters();
        void onEnter();
        void onExit();
        static SingleTypeVariantList GetPreshapeSet(const SingleTypeVariantListPtr preshapes, const std::string& preshapePrefix);

        ConditionIdentifier cond;
        ActionEventIdentifier action;
    };

    DEFINEEVENT(EvReachedIntermediatePreshape)
    DEFINEEVENT(EvReachedFinalPreshape)
    struct SelectAndDoPreshape : StateTemplate<SelectAndDoPreshape>
    {
        void defineSubstates();
        void defineParameters();
        void onEnter();
    };

    struct DoPreshapeSet : StateTemplate<DoPreshapeSet>
    {
        void defineSubstates();
        void defineParameters();
        void onEnter();
        void onExit();

    };





    struct DoPrefixPreshapeSet : StateTemplate<DoPrefixPreshapeSet>
    {
        void defineSubstates();
        void defineParameters();
        void onEnter();
    };





    /**
    *  \ingroup MotionControl
    * Opens the given hand of the robot
    * \param useLeftHand True if left hand should be opened, false for the right hand
    */
    struct OpenHand : StateTemplate<OpenHand>
    {
        void defineSubstates();
        void defineParameters();
        void onEnter();
    };


    /**
    *  \ingroup MotionControl
    * Closes the given hand of the robot
    * \param useLeftHand True if left hand should be closed, false for the right hand
    */
    struct CloseHand : StateTemplate<CloseHand>
    {
        void defineSubstates();
        void defineParameters();
        void onEnter();
    };




    /**
    *  \ingroup MotionControl
    * Moves the head so that it looks at the specified target position
    * \param target Position at which the robot will try to look
    * \param headIKKinematicChainName Name of the kinematic chain for the head that will be sued in the IK
    */
    struct HeadLookAtTarget : StateTemplate<HeadLookAtTarget>
    {
        void defineParameters();
        void defineSubstates();
    };




    /////////////////////////////////////////////////////////////////////
    //// intermediate states
    /////////////////////////////////////////////////////////////////////

    struct CalculateJointAngleConfiguration : StateTemplate<CalculateJointAngleConfiguration>
    {

        ActionEventIdentifier timeoutEvent;
        void defineParameters();
        void run();

    };

    struct ValidateJointAngleConfiguration : StateTemplate<ValidateJointAngleConfiguration>
    {
        RobotStatechartContext* context;
        ActionEventIdentifier timeoutEvent;
        void defineParameters();
        void onEnter();
    };




    struct MoveTCPTrajectoryInit : StateTemplate<MoveTCPTrajectoryInit>
    {
        void defineParameters();
        void onEnter();
    };




    struct MoveTCPTrajectoryCheckCounter : StateTemplate<MoveTCPTrajectoryCheckCounter>
    {
        RobotStatechartContext* context;
        void defineParameters();
        void onEnter();
    };




    struct MoveTCPTrajectoryNextPoint : StateTemplate<MoveTCPTrajectoryNextPoint>
    {
        RobotStatechartContext* context;
        void defineParameters();
        void defineSubstates();
        void onEnter();
    };




    struct CalculateHeadIK : StateTemplate<CalculateHeadIK>
    {
        void defineParameters();
        void onEnter();
        void run();
        FramedPositionPtr targetPosition;
        std::string kinematicChainName;
        std::string virtualPrismaticJointName;
    };

} // namespace MotionControl

} // namespace armarx

#endif // MOTIONCONTROL_H
