/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotStateComponent::
 * @author     ( stefan dot ulbrich at kit dot edu)
 * @date
 * @copyright  http://www.gnu.org/licenses/gpl.txt
 *             GNU General Public License
 */

#ifndef _ARMARX_COMPONENT_RobotStateComponent_RobotStateComponent_H
#define _ARMARX_COMPONENT_RobotStateComponent_RobotStateComponent_H

#include "SharedRobotServants.h"

#include <Core/core/Component.h>
#include <Core/core/application/properties/Properties.h>
#include <Core/core/system/ImportExportComponent.h>
#include <RobotAPI/interface/robotstate/RobotState.h>

#include "remote/RobotStateObjectFactories.h"

namespace armarx
{
    /**
     * RobotStatePropertyDefinition Property Definitions
     */
    class RobotStatePropertyDefinitions:
            public ComponentPropertyDefinitions
    {
    public:
        RobotStatePropertyDefinitions(std::string prefix):
            ComponentPropertyDefinitions(prefix)
        {
            defineRequiredProperty<std::string>("RobotNodeSetName","Set of nodes that is controlled by the KinematicUnit");
            defineRequiredProperty<std::string>("RobotFileName", "Filename of VirtualRobot robot model (e.g. robot_model.xml)");
        }
    };


    /**
     * The RobotStateComponent creates an internal robot representation from
     * a VirtualRobot XML file.
     * Upon request, an Ice proxy to a shared instance of this internal robot
     * can be acquired through a call to RobotStateComponent::getSynchronizedRobot().
     * Additionally it is possible to retrieve a proxy for robot snapshot with
     * RobotStateComponent::getRobotSnapshot().
     * While the synchronized robot will constantly update its internal state
     * the robot snapshot is a clone of the original robot an won't update its
     * configuration over time.
     */
    class ARMARXCOMPONENT_IMPORT_EXPORT RobotStateComponent :
        virtual public Component,
        virtual public RobotStateComponentInterface
    {
    public:
        // inherited from KinematicUnitInterface
        void reportControlModeChanged(const NameControlModeMap& jointModes, bool aValueChanged, const Ice::Current& c = ::Ice::Current());
        void reportJointAngles(const NameValueMap& jointAngles,  bool aValueChanged, const Ice::Current& c = ::Ice::Current());
        void reportJointVelocities(const NameValueMap& jointVelocities,  bool aValueChanged, const Ice::Current& c = ::Ice::Current());
        void reportJointTorques(const NameValueMap& jointTorques,  bool aValueChanged, const Ice::Current& c = ::Ice::Current());
        void reportJointCurrents(const NameValueMap& jointCurrents,  bool aValueChanged, const Ice::Current& c = ::Ice::Current());
        void reportJointMotorTemperatures(const NameValueMap& jointMotorTemperatures,  bool aValueChanged, const Ice::Current& c = ::Ice::Current());
        void reportJointStatuses(const NameStatusMap& jointStatuses,  bool aValueChanged, const Ice::Current& c = ::Ice::Current());

        // inherited from PlatformUnitInterface
        void reportPlatformPose(float currentPlatformPositionX, float currentPlatformPositionY, float currentPlatformRotation, const Ice::Current& c = ::Ice::Current());
        void reportNewTargetPose(float newPlatformPositionX, float newPlatformPositionY, float newPlatformRotation, const Ice::Current& c = ::Ice::Current());
        void reportPlatformVelocity(float currentPlatformVelocityX, float currentPlatformVelocityY, float currentPlatformVelocityRotation, const Ice::Current& c = ::Ice::Current());

        /**
         * \return SharedRobotInterface proxy to the internal synchronized robot (gets created upon first call)
         */
        virtual SharedRobotInterfacePrx getSynchronizedRobot(const Ice::Current&);

        /**
         * \return clone of the internal synchronized robot
         */
        virtual SharedRobotInterfacePrx getRobotSnapshot(const std::string & time, const Ice::Current&);


        /**
         * Create an instance of RobotStatePropertyDefinitions.
         */
        virtual PropertyDefinitionsPtr createPropertyDefinitions();

        virtual std::string getDefaultName() const
        {
            return "RobotStateComponent";
        }
    protected:
        /**
         * Load and create a VirtualRobot::Robot instance from the RobotFileName
         * property. Additionally listen on the KinematicUnit topic for the
         * RobotNodeSet specified in the RobotNodeSetName property.
         */
        virtual void onInitComponent();
        /**
         * Setup RobotStateObjectFactories needed for creating RemoteRobot instances.
         */
        virtual void onConnectComponent();
        /**
         * Calls unref() on RobotStateComponent::_synchronizedPrx.
         */
        virtual ~RobotStateComponent();
    private:
        SharedRobotInterfacePrx _synchronizedPrx;
        VirtualRobot::RobotPtr _synchronized;
    };

}

#endif
