/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotStateComponent::
 * @author     ( stefan dot ulbrich at kit dot edu)
 * @date
 * @copyright  http://www.gnu.org/licenses/gpl.txt
 *             GNU General Public License
 */

#ifndef _ARMARX_COMPONENT_RobotStateComponent_ArmarPose_H
#define _ARMARX_COMPONENT_RobotStateComponent_ArmarPose_H
#include <sstream>

#include <Eigen/Core>
#include <Eigen/Geometry>
#include <RobotAPI/interface/robotstate/RobotState.h>
#include <Core/observers/variant/Variant.h>
#include <Core/observers/AbstractObjectSerializer.h>
#include <VirtualRobot/LinkedCoordinate.h>
#include <VirtualRobot/LinkedCoordinate.h>
#include <VirtualRobot/VirtualRobot.h>
#include <Core/core/exceptions/local/ExpressionException.h>

namespace armarx
{
    namespace VariantType
    {
        // variant types
        const VariantTypeId Vector3 = Variant::addTypeName("::armarx::Vector3Base");
        const VariantTypeId Quaternion = Variant::addTypeName("::armarx::QuaternionBase");
        const VariantTypeId Pose = Variant::addTypeName("::armarx::PoseBase");
        const VariantTypeId FramedVector3 = Variant::addTypeName("::armarx::FramedVector3Base");
        const VariantTypeId FramedPose = Variant::addTypeName("::armarx::FramedPoseBase");
        const VariantTypeId FramedPosition = Variant::addTypeName("::armarx::FramedPositionBase");
        const VariantTypeId FramedOrientation = Variant::addTypeName("::armarx::FramedOrientationBase");
    }


    /**
     * @class Vector3
     * @ingroup Variants
     * @brief The Vector3 class
     */
    class Vector3 :
            virtual public Vector3Base
    {
    public:
        Vector3();
        Vector3(const Eigen::Vector3f &);
        Vector3(const Eigen::Matrix4f &);

        virtual Eigen::Vector3f toEigen() const;

        // inherited from VariantDataClass
        Ice::ObjectPtr ice_clone() const
        {
            return this->clone();
        }
        VariantDataClassPtr clone(const Ice::Current& c = ::Ice::Current()) const
        {
            return new Vector3(*this);
        }
        std::string output(const Ice::Current& c = ::Ice::Current()) const
        {
            std::stringstream s;
            s << toEigen();
            return s.str();
        }
        VariantTypeId getType(const Ice::Current& c = ::Ice::Current()) const
        {
            return VariantType::Vector3;
        }
        bool validate(const Ice::Current& c = ::Ice::Current())
        {
            return true;
        }

        /**
             * @brief Implementation of virtual function to read a Vector3 from an XML-file.
             *Example xml-layout:
             * @code
             *      <x>0</x>
             *      <y>0</y>
             *      <z>0</z>
             * @endcode

             * @param xmlData String with xml-data. NOT a file path!
             * @return ErrorCode, 1 on Success
             */
        int readFromXML(const std::string &xmlData, const Ice::Current &c = ::Ice::Current());
        std::string writeAsXML(const Ice::Current &c = ::Ice::Current());

        friend std::ostream &operator<<(std::ostream &stream, const Vector3& rhs)
        {
            stream << "Vector3: " << std::endl << rhs.output() << std::endl;
            return stream;
        };

    public: // serialization
        virtual void serialize(const armarx::ObjectSerializerBasePtr& serializer, const ::Ice::Current& = ::Ice::Current()) const;
        virtual void deserialize(const armarx::ObjectSerializerBasePtr& serializer, const ::Ice::Current& = ::Ice::Current());
    };

    typedef IceInternal::Handle<Vector3> Vector3Ptr;


    /**
     * @class Quaternion
     * @ingroup Variants
     * @brief The Quaternion class
     */
    class Quaternion :
            virtual public QuaternionBase
    {
    public:
        Quaternion();
        Quaternion(const Eigen::Matrix4f &);
        Quaternion(const Eigen::Matrix3f &);
        Quaternion(const Eigen::Quaternionf &);

        Eigen::Matrix3f toEigen() const;
        Eigen::Matrix3f slerp(float, const Eigen::Matrix3f &);

        static Eigen::Matrix3f slerp(float, const Eigen::Matrix3f &, const Eigen::Matrix3f &);

        // inherited from VariantDataClass
        Ice::ObjectPtr ice_clone() const
        {
            return this->clone();
        }
        VariantDataClassPtr clone(const Ice::Current& c = ::Ice::Current()) const
        {
            return new Quaternion(*this);
        }
        std::string output(const Ice::Current& c = ::Ice::Current()) const
        {
            std::stringstream s;
            s << toEigen();
            return s.str();
        }
        VariantTypeId getType(const Ice::Current& c = ::Ice::Current()) const
        {
            return VariantType::Quaternion;
        }
        bool validate(const Ice::Current& c = ::Ice::Current())
        {
            return true;
        }

        /**
             * @brief Implementation of virtual function to read a Quaternion from an XML-file.
             *Example xml-layout:
             * @code
             *      <qw>0</qw>
             *      <qx>0</qx>
             *      <qy>0</qy>
             *      <qz>0</qz>
             * @endcode
             *
             *OR in AngleAxis-Notation with radiant
             *
             * @code
             *      <angle>0.5</qw>
             *      <x>0.2</x>
             *      <y>0.1</y>
             *      <z>0.7</z>
             * @endcode
             *
             * @param xmlData String with xml-data. NOT a file path!
             * @param c
             * @return ErrorCode, 1 on Success
             */
        int readFromXML(const std::string &xmlData, const Ice::Current &c = ::Ice::Current());
        std::string writeAsXML(const Ice::Current &c = ::Ice::Current());

        friend std::ostream &operator<<(std::ostream &stream, const Quaternion& rhs)
        {
            stream << "Quaternion: " << std::endl << rhs.output() << std::endl;
            return stream;
        };

    public: // serialization
        virtual void serialize(const armarx::ObjectSerializerBasePtr& serializer, const ::Ice::Current& = ::Ice::Current()) const;
        virtual void deserialize(const armarx::ObjectSerializerBasePtr& serializer, const ::Ice::Current& = ::Ice::Current());

    private:
        void init(const Eigen::Matrix3f &);
        void init(const Eigen::Quaternionf &);
    };

    typedef IceInternal::Handle<Quaternion> QuaternionPtr;


    /**
     * @class Pose
     * @ingroup Variants
     * @brief The Pose class
     */
    class Pose :
            virtual public PoseBase
    {
    public:
        Pose();
        Pose(const Eigen::Matrix4f &);
        Pose(const Eigen::Matrix3f &, const Eigen::Vector3f &);

        virtual Eigen::Matrix4f toEigen() const;

        // inherited from VariantDataClass
        Ice::ObjectPtr ice_clone() const
        {
            return this->clone();
        }
        VariantDataClassPtr clone(const Ice::Current& c = ::Ice::Current()) const
        {
            return new Pose(*this);
        }
        std::string output(const Ice::Current& c = ::Ice::Current()) const
        {
            std::stringstream s;
            s << toEigen();
            return s.str();
        }
        VariantTypeId getType(const Ice::Current& c = ::Ice::Current()) const
        {
            return VariantType::Pose;
        }
        bool validate(const Ice::Current& c = ::Ice::Current())
        {
            return true;
        }

        /**
             * @brief Implementation of virtual function to read a FramedOrientation from an XML-file.
             *Example xml-layout:
             * @code
             *      <frame>leftHand</frame>
             *      <x>0</x>
             *      <y>0</y>
             *      <z>0</z>
             *      <qw>0.45</qw>
             *      <qx>0.2</qx>
             *      <qy>0</qy>
             *      <qz>0</qz>
             * @endcode

             * @param xmlData String with xml-data. NOT a file path!
             * @param c
             * @return ErrorCode, 1 on Success
             * @see Quaternion::readFromXml() for AxisAngle-Notation
             */
        int readFromXML(const std::string &xmlData, const Ice::Current &c = ::Ice::Current());
        std::string writeAsXML(const Ice::Current &c = ::Ice::Current());

        friend std::ostream &operator<<(std::ostream &stream, const Pose& rhs)
        {
            stream << "Pose: " << std::endl << rhs.output() << std::endl;
            return stream;
        };

    public: // serialization
        virtual void serialize(const armarx::ObjectSerializerBasePtr& serializer, const ::Ice::Current& = ::Ice::Current()) const;
        virtual void deserialize(const armarx::ObjectSerializerBasePtr& serializer, const ::Ice::Current& = ::Ice::Current());

    private:
        QuaternionPtr c_orientation;
        Vector3Ptr c_position;
    };

    typedef IceInternal::Handle<Pose> PosePtr;



    /**
     * @class FramedPose
     * @ingroup Variants
     * @brief The FramedPose class
     */
    class FramedPose :
            virtual public FramedPoseBase,
            virtual public Pose
    {
    public:
        FramedPose();
        FramedPose(const Eigen::Matrix3f& m, const Eigen::Vector3f& v, const std::string& frame);
        FramedPose(const Eigen::Matrix4f& m, const std::string& frame);

        std::string getFrame() const;

        // inherited from VariantDataClass
        Ice::ObjectPtr ice_clone() const
        {
            return this->clone();
        }

        VariantDataClassPtr clone(const Ice::Current& c = ::Ice::Current()) const
        {
            return new FramedPose(*this);
        }

        std::string output(const Ice::Current& c = ::Ice::Current()) const
        {
            std::stringstream s;
            s << toEigen() << std::endl << "frame: " << getFrame();
            return s.str();
        }

        VariantTypeId getType(const Ice::Current& c = ::Ice::Current()) const
        {
            return VariantType::FramedPose;
        }

        bool validate(const Ice::Current& c = ::Ice::Current())
        {
            return true;
        }

        /**
             * @brief Implementation of virtual function to read a FramedPose from an XML-file.
             *Example xml-layout:
             * @code
             *      <frame>leftHand</frame>
             *      <x>0</x>
             *      <y>0</y>
             *      <z>0</z>
             * @endcode

             * @param xmlData String with xml-data. NOT a file path!
             * @return ErrorCode, 1 on Success
             */
        int readFromXML(const std::string &xmlData, const Ice::Current &c = ::Ice::Current());
        std::string writeAsXML(const Ice::Current &c = ::Ice::Current());

        friend std::ostream &operator<<(std::ostream &stream, const FramedPose& rhs)
        {
            stream << "FramedPose: " << std::endl << rhs.output() << std::endl;
            return stream;
        };

    public:
        virtual void serialize(const armarx::ObjectSerializerBasePtr& serializer, const ::Ice::Current& = ::Ice::Current()) const;
        virtual void deserialize(const armarx::ObjectSerializerBasePtr& serializer, const ::Ice::Current& = ::Ice::Current());
    };

    typedef IceInternal::Handle<FramedPose> FramedPosePtr;


    /**
     * @class FramedVector3
     * @brief FramedVector3 is a 3 dimensional vector with a reference frame.
     * @ingroup Variants
     * The reference frame can be used to change the coordinate system to which
     * the vector relates. The difference to a framed position is, that on frame
     * changing only the orientation of the vector is changed. The translation
     * part remains unchanged. This class is usefull e.g. for forces, torques and tcp
     * velocities.
     *
     * @see Vector3, FramedPosition
     */
    class FramedVector3;
    typedef IceInternal::Handle<FramedVector3> FramedVector3Ptr;

    class FramedVector3 :
            virtual public FramedVector3Base,
            virtual public Vector3
    {
    public:
        FramedVector3();
        FramedVector3(const FramedVector3& source);
        FramedVector3(const Eigen::Vector3f & vec, const std::string &frame );

        std::string getFrame() const;
        static FramedVector3Ptr ChangeFrame(const VirtualRobot::RobotPtr robot, const FramedVector3& framedVec, const std::string &newFrame);


        // inherited from VariantDataClass
        Ice::ObjectPtr ice_clone() const
        {
            return this->clone();
        }
        VariantDataClassPtr clone(const Ice::Current& c = ::Ice::Current()) const
        {
            return new FramedVector3(*this);
        }
        std::string output(const Ice::Current& c = ::Ice::Current()) const
        {
            std::stringstream s;
            s << toEigen() << std::endl << "frame: " << getFrame();
            return s.str();
        }
        VariantTypeId getType(const Ice::Current& c = ::Ice::Current()) const
        {
            return VariantType::FramedVector3;
        }
        bool validate(const Ice::Current& c = ::Ice::Current())
        {
            return true;
        }

        /**
             * @brief Implementation of virtual function to read a FramedVector3 from an XML-file.
             *Example xml-layout:
             * @code
             *      <frame>leftHand</frame>
             *      <x>0</x>
             *      <y>0</y>
             *      <z>0</z>
             * @endcode

             * @param xmlData String with xml-data. NOT a file path!
             * @return ErrorCode, 1 on Success
             */
        int readFromXML(const std::string &xmlData, const Ice::Current &c = ::Ice::Current());
        std::string writeAsXML(const Ice::Current &c = ::Ice::Current());

        friend std::ostream &operator<<(std::ostream &stream, const FramedVector3& rhs)
        {
            stream << "FramedVector3: " << std::endl << rhs.output() << std::endl;
            return stream;
        }

    public: // serialization
        virtual void serialize(const armarx::ObjectSerializerBasePtr& serializer, const ::Ice::Current& = ::Ice::Current()) const;
        virtual void deserialize(const armarx::ObjectSerializerBasePtr& serializer, const ::Ice::Current& = ::Ice::Current());
    private:

        static Eigen::Matrix4f __GetRotationBetweenFrames(const std::string &oldFrame, const std::string &newFrame, VirtualRobot::RobotPtr robotState);
    };



    /**
     * @class FramedPosition
     * @ingroup Variants
     * @brief The FramedPosition class
     */
    class FramedPosition :
            virtual public FramedPositionBase,
            virtual public Vector3
    {
    public:
        FramedPosition( );
        FramedPosition(const Eigen::Vector3f &, const std::string &frame );
        FramedPosition(const Eigen::Matrix4f &, const std::string &frame );
        //FramedPosition(const Vector3BasePtr pos, const std::string &frame ); // this doesnt work for unknown reasons
        std::string getFrame() const;

        // inherited from VariantDataClass
        Ice::ObjectPtr ice_clone() const
        {
            return this->clone();
        }
        VariantDataClassPtr clone(const Ice::Current& c = ::Ice::Current()) const
        {
            return new FramedPosition(*this);
        }
        std::string output(const Ice::Current& c = ::Ice::Current()) const
        {
            std::stringstream s;
            s << toEigen() << std::endl << "frame: " << getFrame();
            return s.str();
        }
        VariantTypeId getType(const Ice::Current& c = ::Ice::Current()) const
        {
            return VariantType::FramedPosition;
        }
        bool validate(const Ice::Current& c = ::Ice::Current())
        {
            return true;
        }

        /**
             * @brief Implementation of virtual function to read a FramedPosition from an XML-file.
             *Example xml-layout:
             * @code
             *      <frame>leftHand</frame>
             *      <x>0</x>
             *      <y>0</y>
             *      <z>0</z>
             * @endcode

             * @param xmlData String with xml-data. NOT a file path!
             * @return ErrorCode, 1 on Success
             */
        int readFromXML(const std::string &xmlData, const Ice::Current &c = ::Ice::Current());
        std::string writeAsXML(const Ice::Current &c = ::Ice::Current());

        friend std::ostream &operator<<(std::ostream &stream, const FramedPosition& rhs)
        {
            stream << "FramedPosition: " << std::endl << rhs.output() << std::endl;
            return stream;
        };

    public: // serialization
        virtual void serialize(const armarx::ObjectSerializerBasePtr& serializer, const ::Ice::Current& = ::Ice::Current()) const;
        virtual void deserialize(const armarx::ObjectSerializerBasePtr& serializer, const ::Ice::Current& = ::Ice::Current());
    };

    typedef IceInternal::Handle<FramedPosition> FramedPositionPtr;


    /**
     * @class FramedOrientation
     * @ingroup Variants
     * @brief The FramedOrientation class
     */
    class FramedOrientation :
            virtual public FramedOrientationBase,
            virtual public Quaternion
    {
    public:
        FramedOrientation();
        FramedOrientation(const Eigen::Matrix4f &, const std::string &frame );
        FramedOrientation(const Eigen::Matrix3f &, const std::string &frame );
        // this doesnt work for an unknown reason
        //FramedOrientation(const QuaternionBasePtr ori, const std::string &frame );
        std::string getFrame()const ;

        // inherited from VariantDataClass
        Ice::ObjectPtr ice_clone() const
        {
            return this->clone();
        }
        VariantDataClassPtr clone(const Ice::Current& c = ::Ice::Current()) const
        {
            return new FramedOrientation(*this);
        }
        std::string output(const Ice::Current& c = ::Ice::Current()) const
        {
            std::stringstream s;
            s << toEigen() << std::endl << "frame: " << getFrame();
            return s.str();
        }
        VariantTypeId getType(const Ice::Current& c = ::Ice::Current()) const
        {
            return VariantType::FramedOrientation;
        }
        bool validate(const Ice::Current& c = ::Ice::Current())
        {
            return true;
        }

        /**
             * @brief Implementation of virtual function to read a FramedOrientation from an XML-file.
             *Example xml-layout:
             * @code
             *      <frame>leftHand</frame>
             *      <qw>0.45</qw>
             *      <qx>0.2</qx>
             *      <qy>0</qy>
             *      <qz>0</qz>
             * @endcode

             * @param xmlData String with xml-data. NOT a file path!
             * @param c
             * @return ErrorCode, 1 on Success
             */
        int readFromXML(const std::string &xmlData, const Ice::Current &c = ::Ice::Current());
        std::string writeAsXML(const Ice::Current &c = ::Ice::Current());

        friend std::ostream &operator<<(std::ostream &stream, const FramedOrientation& rhs)
        {
            stream << "FramedOrientation: " << std::endl << rhs.output() << std::endl;
            return stream;
        };

    public: // serialization
        virtual void serialize(const armarx::ObjectSerializerBasePtr& serializer, const ::Ice::Current& = ::Ice::Current()) const;
        virtual void deserialize(const armarx::ObjectSerializerBasePtr& serializer, const ::Ice::Current& = ::Ice::Current());
    };

    typedef IceInternal::Handle<FramedOrientation> FramedOrientationPtr;


    namespace ArmarPose
    {
        VirtualRobot::LinkedCoordinate createLinkedCoordinate(VirtualRobot::RobotPtr &virtualRobot, const FramedPositionPtr& position, const FramedOrientationPtr& orientation);
    }
}
#endif
