/*
* This file is part of ArmarX.
*
* ArmarX is free software; you can redistribute it and/or modify
* it under the terms of the GNU Lesser General Public License as
* published by the Free Software Foundation; either version 2 of
* the License, or (at your option) any later version.
*
* ArmarX is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU Lesser General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*
* @package    RobotAPI::GraspingWithTorques
* @author     Markus Przybylski
* @date       2014 Markus Przybylski

* @copyright  http://www.gnu.org/licenses/gpl.txt
*             GNU General Public License
*/

#include "GraspingWithTorques.h"
//#include "Armar3GraspContext.h"   //MP: Maybe necessary again later?
#include "../../core/RobotStatechartContext.h"

#include <Core/observers/variant/ChannelRef.h>
#include <Core/observers/variant/SingleTypeVariantList.h>
#include <RobotAPI/robotstate/remote/ArmarPose.h>
#include <VirtualRobot/IK/DifferentialIK.h>

#include <RobotAPI/interface/units/KinematicUnitInterface.h>
#include <RobotAPI/interface/units/HandUnitInterface.h>

#include <RobotAPI/core/RobotStatechartContext.h>

namespace armarx
{
    // ****************************************************************
    // Implementation of StatechartGraspingWithTorques
    // ****************************************************************
    void StatechartGraspingWithTorques::defineParameters()
    {
        //preshape settings:
        addToInput("jointAnglesPreshape", VariantType::List(VariantType::Float), false);
        //how long to wait before switching from preshape to grasp
        addToInput("timeoutPreshape", VariantType::Float, false);
        addToInput("useTorquesForGrasping", VariantType::Int, false);

        //closeHandWithTorques settings:
        addToInput("jointTorquesGrasp", VariantType::List(VariantType::Float), false);        
        //Make sure that the joints have time to start moving before we check
        //low joint velocities as a termination criterion.
        addToInput("timeoutMinExecutionTime", VariantType::Float, false);

        //closeHandWithJointAngles settings:
        addToInput("jointAnglesGrasp", VariantType::List(VariantType::Float), false);

        //installTerminateConditions settings:
        //how long to wait until you think the hand is closed and grasping is completed
        addToInput("timeoutGrasp", VariantType::Float, false);
        addToInput("thresholdVelocity", VariantType::Float, false);
        addToInput("handUnitName", VariantType::String, false);


        addToInput("robotNodeSetName", VariantType::String, false);

        addToLocal("jointNames", VariantType::List(VariantType::String));

        //addToLocal("jointVelocityChannel", VariantType::ChannelRef);  //first try ...

        addToLocal("jointVelocitiesDatafields", VariantType::List(VariantType::DatafieldRef));

    }

    void StatechartGraspingWithTorques::defineSubstates()
    {
        StatePtr statePreshape = addState<StatePreshape>("statePreshape");
        StatePtr stateCloseHandWithTorques = addState<StateCloseHandWithTorques>("stateCloseHandWithTorques");
        StatePtr stateCloseHandWithJointAngles = addState<StateCloseHandWithJointAngles>("stateCloseHandWithJointAngles");
        StatePtr stateInstallTerminateConditions = addState<StateInstallTerminateConditions>("stateInstallTerminateConditions");
        StatePtr stateSuccess = addState<SuccessState>("stateSuccess");
        StatePtr stateFailure = addState<FailureState>("stateFailure");

        ParameterMappingPtr mapPreshapeInfo = ParameterMapping::createMapping()
                ->mapFromParent("jointAnglesPreshape", "jointAnglesPreshape")
                ->mapFromParent("timeoutPreshape", "timeoutPreshape")
                ->mapFromParent("jointNames", "jointNames")
                ->mapFromParent("useTorquesForGrasping", "useTorquesForGrasping");

        ParameterMappingPtr mapCloseHandWithTorquesInfo = ParameterMapping::createMapping()
                ->mapFromParent("jointTorquesGrasp", "jointTorquesGrasp")
                ->mapFromParent("timeoutMinExecutionTime", "timeoutMinExecutionTime")
                ->mapFromParent("jointNames", "jointNames");

        ParameterMappingPtr mapCloseHandWithJointAnglesInfo = ParameterMapping::createMapping()
                ->mapFromParent("jointAnglesGrasp", "jointAnglesGrasp")
                ->mapFromParent("timeoutGrasp", "timeoutGrasp")
                ->mapFromParent("jointNames", "jointNames");

        ParameterMappingPtr mapInstallTerminateConditionsInfo = ParameterMapping::createMapping()
                ->mapFromParent("timeoutGrasp", "timeoutGrasp")
                ->mapFromParent("thresholdVelocity", "thresholdVelocity")
                ->mapFromParent("jointNames")
                ->mapFromParent("handUnitName", "handUnitName")
                ->mapFromParent("jointVelocitiesDatafields");
                //->mapFromParent("jointVelocityChannel"); //first try...

        setInitState(statePreshape, mapPreshapeInfo);

        //transitions
        addTransition<EvPreshapeTimeout_ToCloseWithTorques>(statePreshape, stateCloseHandWithTorques, mapCloseHandWithTorquesInfo);
        addTransition<EvPreshapeTimeout_ToCloseWithJointAngles>(statePreshape, stateCloseHandWithJointAngles, mapCloseHandWithJointAnglesInfo);
        addTransition<EvMinExecutionTimeout>(stateCloseHandWithTorques, stateInstallTerminateConditions, mapInstallTerminateConditionsInfo);
        //addTransition<EvGraspWithTorquesTimeout>(stateInstallTerminateConditions, stateSuccess);
        addTransition<EvMinExecutionTimeout>(stateInstallTerminateConditions, stateSuccess); // timeout is ok for now....
        addTransition<EvAllJointVelocitiesLow>(stateInstallTerminateConditions, stateSuccess);
        addTransition<EvGraspWithJointAnglesTimeout>(stateCloseHandWithJointAngles, stateSuccess);

    }

    void StatechartGraspingWithTorques::onEnter()
    {
        ARMARX_LOG << eINFO << "Entering StatechartGraspingWithTorques";

        RobotStatechartContext* context = getContext<RobotStatechartContext>();
        //setLocal("jointVelocityChannel", context->getChannelRef(context->getKinematicUnitObserverName(),"jointvelocities"));
        ChannelRefPtr tempChannelRef = context->getChannelRef(context->getKinematicUnitObserverName(),"jointvelocities");

        VirtualRobot::RobotNodeSetPtr nodeSet = context->remoteRobot->getRobotNodeSet(getInput<std::string>("robotNodeSetName"));    //woher robotNodeSetName holen? (Argument für getRobotNodeSet())

        SingleTypeVariantList jointNames(VariantType::String);
        SingleTypeVariantList dataFields(VariantType::DatafieldRef);


        for (size_t i=0; i<nodeSet->getSize(); i++)
        {
            jointNames.addVariant(nodeSet->getNode(i)->getName());      //nodes = joints
            dataFields.addVariant(context->getDatafieldRef(tempChannelRef, nodeSet->getNode(i)->getName()));
        }

        setLocal("jointNames", jointNames);
        setLocal("jointVelocitiesDatafields", dataFields);

    }

    void StatechartGraspingWithTorques::onExit()
    {

        ARMARX_LOG << eINFO << "Exiting StatechartGraspingWithTorques...";

    }

    // ****************************************************************
    // Implementation of StatePreshape
    // ****************************************************************

    void StatePreshape::defineParameters()
    {
        addToInput("jointAnglesPreshape", VariantType::List(VariantType::Float), false);
        addToInput("timeoutPreshape", VariantType::Float, false);

        //addToLocal("jointNames", VariantType::List(VariantType::String));
        addToInput("jointNames", VariantType::List(VariantType::String), false);
        addToInput("useTorquesForGrasping", VariantType::Int, false);
    }

    void StatePreshape::onEnter()
    {
        ARMARX_LOG << eIMPORTANT << "Entering statePreshape";

        RobotStatechartContext* rsContext = getContext<RobotStatechartContext>();
        SingleTypeVariantListPtr jointNames = getInput<SingleTypeVariantList>("jointNames");
        SingleTypeVariantListPtr jointAnglesPreshapeList = getInput<SingleTypeVariantList>("jointAnglesPreshape");
        NameValueMap jointNamesAndValues;

        if (jointNames->getSize() == jointAnglesPreshapeList->getSize())
        {
            for (int j=0; j<jointNames->getSize(); j++)
            {
                jointNamesAndValues[jointNames->getVariant(j)->getString()]=jointAnglesPreshapeList->getVariant(j)->getFloat();
            }
        }
        else
            throw LocalException("StatePreshape: List lengths do not match!");

        rsContext->kinematicUnitPrx->setJointAngles(jointNamesAndValues);

        ARMARX_LOG << eINFO << "Installing preshape timeout condition";
        float timeoutPreshape = getInput<float>("timeoutPreshape");

        int useTorquesForGrasping = getInput<int>("useTorquesForGrasping");
        if (useTorquesForGrasping > 0)
            condPreshapeTimeout = setTimeoutEvent(timeoutPreshape, createEvent<EvPreshapeTimeout_ToCloseWithTorques>());
        else
            condPreshapeTimeout = setTimeoutEvent(timeoutPreshape, createEvent<EvPreshapeTimeout_ToCloseWithJointAngles>());

        ARMARX_LOG << eINFO << "Done StatePreshape::onEnter()";
    }

    void StatePreshape::onExit()
    {
        removeTimeoutEvent(condPreshapeTimeout);
        ARMARX_LOG << eIMPORTANT << "StatePreshape: Done onExit()";
    }

    // ****************************************************************
    // Implementation of StateCloseHandWithTorques
    // ****************************************************************

    void StateCloseHandWithTorques::defineParameters()
    {
        //closeHandWithTorques settings:
        addToInput("jointTorquesGrasp", VariantType::List(VariantType::Float), false);
        addToInput("timeoutMinExecutionTime", VariantType::Float, false);

        addToInput("jointNames", VariantType::List(VariantType::String), false);
    }

    void StateCloseHandWithTorques::onEnter()
    {
        ARMARX_LOG << eIMPORTANT << "Entering stateCloseHandWithTorques";

        RobotStatechartContext* rsContext = getContext<RobotStatechartContext>();
        SingleTypeVariantListPtr jointNames = getInput<SingleTypeVariantList>("jointNames");
        SingleTypeVariantListPtr jointTorquesGraspList = getInput<SingleTypeVariantList>("jointTorquesGrasp");
        NameValueMap jointNamesAndValues;

        if (jointNames->getSize() == jointTorquesGraspList->getSize())
        {
            for (int j=0; j<jointNames->getSize(); j++)
            {
                jointNamesAndValues[jointNames->getVariant(j)->getString()]=jointTorquesGraspList->getVariant(j)->getFloat();
            }
        }
        else
            throw LocalException("StateCloseHandWithTorques: List lengths do not match!");

        rsContext->kinematicUnitPrx->setJointTorques(jointNamesAndValues);

        ARMARX_LOG << eINFO << "Installing timeoutMinExecutionTime condition";
        float timeoutMinExecutionTime = getInput<float>("timeoutMinExecutionTime");
        condMinimumExecutionTimeout = setTimeoutEvent(timeoutMinExecutionTime, createEvent<EvMinExecutionTimeout>());

        ARMARX_LOG << eIMPORTANT << "Done StateCloseHandWithTorques::onEnter()";
    }

    void StateCloseHandWithTorques::onExit()
    {
        removeTimeoutEvent(condMinimumExecutionTimeout);
    }

    // ****************************************************************
    // Implementation of StateCloseHandWithJointAngles
    // ****************************************************************

    void StateCloseHandWithJointAngles::defineParameters()
    {
        //closeHandWithJointAngles settings:
        addToInput("jointAnglesGrasp", VariantType::List(VariantType::Float), false);
        addToInput("timeoutGrasp", VariantType::Float, false);

        addToInput("jointNames", VariantType::List(VariantType::String), false);
    }

    void StateCloseHandWithJointAngles::onEnter()
    {
        ARMARX_LOG << eIMPORTANT << "Entering stateCloseHandWithJointAngles !!!";

        RobotStatechartContext* rsContext = getContext<RobotStatechartContext>();
        SingleTypeVariantListPtr jointNames = getInput<SingleTypeVariantList>("jointNames");
        SingleTypeVariantListPtr jointAnglesGraspList = getInput<SingleTypeVariantList>("jointAnglesGrasp");
        NameValueMap jointNamesAndValues;
        NameControlModeMap controlModes;

        if (jointNames->getSize() == jointAnglesGraspList->getSize())
        {
            for (int j=0; j<jointNames->getSize(); j++)
            {
                jointNamesAndValues[jointNames->getVariant(j)->getString()]=jointAnglesGraspList->getVariant(j)->getFloat();
                controlModes[jointNames->getVariant(j)->getString()] = ePositionControl;
            }
        }
        else
            throw LocalException("stateCloseHandWithJointAngles: List lengths do not match!");

        rsContext->kinematicUnitPrx->switchControlMode(controlModes);
        rsContext->kinematicUnitPrx->setJointAngles(jointNamesAndValues);

        ARMARX_LOG << eINFO << "Installing timeoutGrasp condition";
        float timeoutGrasp = getInput<float>("timeoutGrasp");
        condGraspWithJointAnglesTimeout = setTimeoutEvent(timeoutGrasp, createEvent<EvGraspWithJointAnglesTimeout>());

        ARMARX_LOG << eIMPORTANT << "Done B " << BOOST_CURRENT_FUNCTION;
    }

    void StateCloseHandWithJointAngles::onExit()
    {
        removeTimeoutEvent(condGraspWithJointAnglesTimeout);
    }


    // ****************************************************************
    // Implementation of StateInstallTerminateConditions
    // ****************************************************************

    void StateInstallTerminateConditions::defineParameters()
    {
        //installTerminateConditions settings:
        //how long to wait until you think the hand is closed and grasping is completed
        addToInput("timeoutGrasp", VariantType::Float, false);
        addToInput("thresholdVelocity", VariantType::Float, false);

        addToInput("jointNames", VariantType::List(VariantType::String), false);

        //addToInput("jointVelocityChannel", VariantType::ChannelRef, false);   //first try ...

        addToInput("jointVelocitiesDatafields", VariantType::List(VariantType::DatafieldRef), false);
        addToInput("handUnitName", VariantType::String, false);
    }

    void StateInstallTerminateConditions::onEnter()
    {
        ARMARX_LOG << eIMPORTANT << "onEnter() stateInstallTerminateConditions";

        ARMARX_LOG << eINFO << "Installing timeoutGrasp condition";
        float timeoutGrasp = getInput<float>("timeoutGrasp");
        condGraspTimeout = setTimeoutEvent(timeoutGrasp, createEvent<EvMinExecutionTimeout>());

        float thresholdVelocity = getInput<float>("thresholdVelocity");
        ARMARX_LOG << eINFO << "Installing allJointVelocitiesLow condition, threshold: " << thresholdVelocity << flush;

        //=======
        //first try ... (with ChannelRef)
        //=======

        /*
        Term allJointVelocitiesLow;
        SingleTypeVariantListPtr jointNamesList = getInput<SingleTypeVariantList>("jointNames");
        for (int i=0; i<jointNamesList->getSize(); i++)
        {
           allJointVelocitiesLow = allJointVelocitiesLow && Literal(getInput<ChannelRef>("jointVelocityChannel")
                         ->getDataFieldIdentifier(jointNamesList->getVariant(i)->getString()), "smaller", Literal::createParameterList(thresholdVelocity));
        }
        condAllJointVelocitiesLow = installCondition(allJointVelocitiesLow, createEvent<EvAllJointVelocitiesLow>());
        */

        //=======
        //second try ... (with DatafieldRef)
        //=======
        Term allJointVelocitiesLow_NEW;
        //SingleTypeVariantListPtr jointNamesList = getInput<SingleTypeVariantList>("jointNames");
        SingleTypeVariantListPtr dataFieldsList = getInput<SingleTypeVariantList>("jointVelocitiesDatafields");

        //for (int i=0; i<jointNamesList->getSize(); i++)
        for (int i=0; i<dataFieldsList->getSize(); i++)
        {
            allJointVelocitiesLow_NEW = allJointVelocitiesLow_NEW &&
                    Literal(dataFieldsList->getVariant(i)->get<DatafieldRef>()->getDataFieldIdentifier(),
                            "inrange", Literal::createParameterList(-thresholdVelocity, thresholdVelocity));
        }

        condAllJointVelocitiesLow = installCondition(allJointVelocitiesLow_NEW, createEvent<EvAllJointVelocitiesLow>());

    }

    void StateInstallTerminateConditions::onExit()
    {
        ARMARX_LOG << eIMPORTANT << "onExit() stateInstallTerminateConditions";
        removeTimeoutEvent(condGraspTimeout);
        removeCondition(condAllJointVelocitiesLow);

        //---------
        //set joint velocities manually to zero (DEBUG)
        //---------

        std::string handUnitName = getInput<std::string>("handUnitName");
        ARMARX_LOG << eIMPORTANT << "xx..................SENDING OBJECTGRASPED TO hand unit name " << handUnitName << "...................";

        RobotStatechartContext* rsContext = getContext<RobotStatechartContext>();
        if (rsContext->getHandUnit(handUnitName))
        {
            ARMARX_LOG << eIMPORTANT << "xx..................SENDING OBJECTGRASPED TO hand unit name " << handUnitName << ".......OK: sending Vitalis............";
            rsContext->getHandUnit(handUnitName)->setObjectGrasped("Vitalis");
        }

        SingleTypeVariantListPtr jointNames = getInput<SingleTypeVariantList>("jointNames");
        //SingleTypeVariantListPtr jointTorquesGraspList = getInput<SingleTypeVariantList>("jointTorquesGrasp");
        NameValueMap jointNamesAndValues;
        NameControlModeMap controlModes;

        for (int j=0; j<jointNames->getSize(); j++)
        {
            // no, now we set it to zero.... we want to ensure that the object stayes grasped -> apply a small velocity (todo: find a better solution!)
            jointNamesAndValues[jointNames->getVariant(j)->getString()] = 0.0;//1f;  //set everything to zero
            controlModes[jointNames->getVariant(j)->getString()] = eVelocityControl;
        }

        rsContext->kinematicUnitPrx->switchControlMode(controlModes);
//        rsContext->kinematicUnitPrx->setJointTorques(jointNamesAndValues);
        rsContext->kinematicUnitPrx->setJointVelocities(jointNamesAndValues);
    }

}

