/*
* This file is part of ArmarX.
*
* ArmarX is free software; you can redistribute it and/or modify
* it under the terms of the GNU Lesser General Public License as
* published by the Free Software Foundation; either version 2 of
* the License, or (at your option) any later version.
*
* ArmarX is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU Lesser General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*
* @package    RobotAPI::MovePlatform
* @author     Valerij Wittenbeck
* @date       2014

* @copyright  http://www.gnu.org/licenses/gpl.txt
*             GNU General Public License
*/

#include "MovePlatform.h"
//#include "../core/RobotStatechartContext.h"

#include <Core/observers/variant/ChannelRef.h>
#include <Core/observers/variant/SingleTypeVariantList.h>
#include <RobotAPI/robotstate/remote/ArmarPose.h>
#include <VirtualRobot/IK/DifferentialIK.h>

#include <RobotAPI/interface/units/PlatformUnitInterface.h>

namespace armarx
{


    void MovePlatformStateChart::onInitRemoteStateOfferer()
    {
        addState<StatechartMovePlatform>("MovePlatformStateChart");
    }

    void MovePlatformStateChart::onConnectRemoteStateOfferer()
    {

    }

    // ****************************************************************
    // Implementation of StatechartMovePlatform
    // ****************************************************************
    void StatechartMovePlatform::defineParameters()
    {
//        setConfigFile("RobotAPI/scenarios/MovePlatformTest/configs/MovePlatformExample.xml");
        setConfigFile("RobotAPI/scenarios/MovePlatformTest/configs/MovePlatformExampleCloseToTable.xml");
        addToInput("targetPositions", VariantType::List(VariantType::Vector3), false);
        addToInput("positionalAccuracy", VariantType::Float, false);
        addToInput("orientationalAccuracy", VariantType::Float, false);
        addToInput("timeoutMoveTo", VariantType::Int, false);
        addToInput("waitAfterLast", VariantType::Int, false);
        addToLocal("positionCounter", VariantType::ChannelRef);
    }

    void StatechartMovePlatform::defineSubstates()
    {
        ARMARX_LOG << eVERBOSE << "Entering StatechartMovePlatform";
        StatePtr stateMoveToNext = addState<StateMoveToNext>("stateMoveToNext");
        StatePtr stateSuccess = addState<SuccessState>("stateSuccess");
        StatePtr stateFailure = addState<FailureState>("stateFailure");

        ParameterMappingPtr mapMoveToNextInfo = ParameterMapping::createMapping()
                ->mapFromParent("*", "*");
        setInitState(stateMoveToNext, mapMoveToNextInfo);

        //transitions
        addTransition<EvTargetReached>(stateMoveToNext, stateMoveToNext, mapMoveToNextInfo);
        addTransition<EvNoTargets>(stateMoveToNext, stateSuccess);
        addTransition<EvTimeoutTargetUnreachable>(stateMoveToNext, stateFailure);
    }

    void StatechartMovePlatform::onEnter()
    {
        ARMARX_LOG << eVERBOSE << "Entering StatechartMovePlatform";

        PlatformContext* context = getContext<PlatformContext>();
        ChannelRefPtr counter = ChannelRefPtr::dynamicCast(context->systemObserverPrx->startCounter(0, "positionCounter"));

        setLocal("positionCounter", counter);
    }

    void StatechartMovePlatform::onExit()
    {
        ChannelRefPtr r = getLocal<ChannelRef>("positionCounter");
        getContext()->systemObserverPrx->removeCounter(r);
        ARMARX_LOG << eVERBOSE << "Exiting StatechartMovePlatform...";
    }

    // ****************************************************************
    // Implementation of StateMoveToNext
    // ****************************************************************

    void StateMoveToNext::defineParameters()
    {
        addToInput("targetPositions", VariantType::List(VariantType::Vector3), false);
        addToInput("positionalAccuracy", VariantType::Float, false);
        addToInput("orientationalAccuracy", VariantType::Float, false);
        addToInput("timeoutMoveTo", VariantType::Int, false);
        addToInput("waitAfterLast", VariantType::Int, false);
        addToInput("positionCounter", VariantType::ChannelRef, false);
    }

    void StateMoveToNext::onEnter()
    {
        ARMARX_LOG << eVERBOSE << "Entering StateMoveToNext::onEnter" << flush;
        PlatformContext* context = getContext<PlatformContext>();
        ChannelRefPtr counter = getInput<ChannelRef>("positionCounter");
        int positionIndex = counter->getDataField("value")->getInt();
        ARMARX_DEBUG << "Entering positionIndex:" << positionIndex << flush;
        SingleTypeVariantListPtr points = getInput<SingleTypeVariantList>("targetPositions");
        ARMARX_DEBUG << "points->getSize:" << points->getSize() << flush;
        if (positionIndex < points->getSize()) {
            SingleTypeVariantListPtr list = getInput<SingleTypeVariantList>("targetPositions");

            ARMARX_DEBUG << "list->getSize:" << list->getSize() << flush;
            VariantPtr v = list->getVariant(positionIndex);
            ARMARX_DEBUG << "v->getTypeName():" << v->getTypeName() << flush;

            Vector3Ptr currentTarget = v->get<Vector3>();
            ARMARX_DEBUG << "currentTarget:" << currentTarget->toEigen().transpose() << flush;
            float positionalAccuracy = getInput<float>("positionalAccuracy");
            float orientationalAccuracy = getInput<float>("orientationalAccuracy");
            context->platformUnitPrx->moveTo(currentTarget->x, currentTarget->y, currentTarget->z, positionalAccuracy, orientationalAccuracy);

            Literal checkX("PlatformUnitObserver.platformPose.positionX", "inrange", Literal::createParameterList(currentTarget->x-positionalAccuracy,currentTarget->x+positionalAccuracy));
            Literal checkY("PlatformUnitObserver.platformPose.positionY", "inrange", Literal::createParameterList(currentTarget->y-positionalAccuracy,currentTarget->y+positionalAccuracy));
            Literal checkAngle("PlatformUnitObserver.platformPose.rotation", "inrange", Literal::createParameterList(currentTarget->z-orientationalAccuracy,currentTarget->z+orientationalAccuracy));

            condTargetReached = installCondition<EvTargetReached>(checkX && checkY && checkAngle);
            timeoutMoveToNext = setTimeoutEvent(getInput<int>("timeoutMoveTo"), createEvent<EvTimeoutTargetUnreachable>());
        } else {
//            sendEvent<EvNoTargets>();
            timeoutWaitAfterLast = setTimeoutEvent(getInput<int>("waitAfterLast"), createEvent<EvNoTargets>());
        }

        context->systemObserverPrx->incrementCounter(counter);
        ARMARX_LOG << eVERBOSE << "Done MoveToNext::onEnter()";
    }

    void StateMoveToNext::onExit()
    {
        removeCondition(condTargetReached);
        removeTimeoutEvent(timeoutMoveToNext);
        removeTimeoutEvent(timeoutWaitAfterLast);
        ARMARX_LOG << eVERBOSE << "StateMoveToNext: Done onExit()";
    }
}

