/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    ArmarXCore::units
 * @author     Manfred Kroehnert (manfred dot kroehnert at kit dot edu)
 * @date       2013
 * @copyright  http://www.gnu.org/licenses/gpl.txt
 *             GNU General Public License
 */

#ifndef _ARMARX_CORE_UNITS_HANDUNIT_H
#define _ARMARX_CORE_UNITS_HANDUNIT_H

#include <RobotAPI/units/SensorActorUnit.h>

#include <RobotAPI/interface/units/HandUnitInterface.h>

#include <Core/core/application/properties/Properties.h>
#include <Core/core/system/ImportExportComponent.h>
#include <Core/observers/variant/SingleTypeVariantList.h>

#include <VirtualRobot/EndEffector/EndEffector.h>

namespace VirtualRobot
{
    class Robot;
    typedef boost::shared_ptr<Robot> RobotPtr;
}
namespace armarx
{
    /**
     * @class HandUnitPropertyDefinitions
     * @brief Defines all necessary properties for armarx::HandUnit
     * @ingroup SensorActorUnits
     */
    class HandUnitPropertyDefinitions:
            public ComponentPropertyDefinitions
    {
    public:
        HandUnitPropertyDefinitions(std::string prefix):
            ComponentPropertyDefinitions(prefix)
        {
            defineRequiredProperty<std::string>("RobotFileName","VirtualRobot XML file in which the endeffector is is stored.");
            defineRequiredProperty<std::string>("EndeffectorName","Name of the endeffector as stored in the XML file (will publish values on EndeffectorName + 'State')");
        }
    };


    /**
     * @class HandUnit
     * @brief This class defines an interface for providing high level access to robot hands
     * @ingroup SensorActorUnits
     *
     * An instance of a HandUnit provides means to open, close, and preshape hands.
     * It uses the HandUnitListener Ice interface to report updates of its current state
     *
     */
    class HandUnit :
        virtual public HandUnitInterface,
        virtual public SensorActorUnit
    {
    public:
        // inherited from Component
        virtual std::string getDefaultName() const
        {
            return "HandUnit";
        }

        /**
         * Retrieve proxy for publishing State information and call
         * armarx::PlatformUnit::onInitPlatformUnit().
         * @see armarx::Component::onInitComponent()
         */
        virtual void onInitComponent();
        /**
         * Calls armarx::PlatformUnit::onStartPlatformUnit().
         * @see armarx::Component::onConnectComponent()
         */
        virtual void onConnectComponent();
        /**
         * Calls armarx::PlatformUnit::onExitPlatformUnit().
         * @see armarx::Component::onExitComponent()
         */
        virtual void onExitComponent();

        /**
         *
         */
        virtual void onInitHandUnit() = 0;
        /**
         *
         */
        virtual void onStartHandUnit() = 0;
        /**
         *
         */
        virtual void onExitHandUnit() = 0;

        /**
         * Send command to the hand to open all fingers.
         */
        virtual void open(const Ice::Current& c = ::Ice::Current());

        /**
         * Send command to the hand to close all fingers.
         */
        virtual void close(const Ice::Current& c = ::Ice::Current());

        /**
         * Send command to the hand to form a specific preshape position.
         */
        virtual void preshape(const std::string& preshapeName, const Ice::Current& c = ::Ice::Current());

        /**
         * @return a list of strings for preshape positions which can be used together with HandUnit::preshape().
         */
        virtual SingleTypeVariantListBasePtr getPreshapeNames(const Ice::Current& c = ::Ice::Current());

        NameValueMap getPreshapeJointValues(const std::string &preshapeName, const Ice::Current &);

         void setObjectGrasped(const std::string &objectName, const Ice::Current &);
         void setObjectReleased(const std::string &objectName, const Ice::Current &);


        /**
         * @see armarx::PropertyUser::createPropertyDefinitions()
         */
        virtual PropertyDefinitionsPtr createPropertyDefinitions();

    protected:
        /**
         * HandUnitListener proxy for publishing state updates
         */
        HandUnitListenerPrx listenerPrx;
        /**
         * Ice Topic name on which armarx::HandUnit::listenerPrx publishes information
         */
        std::string listenerChannelName;
        /**
         * List containing the names of all valid preshapes.
         */
        SingleTypeVariantListPtr preshapeNames;

        VirtualRobot::RobotPtr robot;
        VirtualRobot::EndEffectorPtr eef;

        std::string robotName;
        std::string tcpName;

        std::map <std::string, float> handJoints;

        std::string graspedObject;

    };
}

#endif
