/*
* This file is part of ArmarX.
*
* ArmarX is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as
* published by the Free Software Foundation; either version 2 of
* the License, or (at your option) any later version.
*
* ArmarX is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU Lesser General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*
* @package    ArmarX::Core
* @author     Kai Welke (welke _at_ kit _dot_ edu)
* @date       2011 Kai Welke
* @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
*             GNU General Public License
*/

#ifndef _ARMARX_CORE_KINEMATIC_UNIT_OBSERVER_H
#define _ARMARX_CORE_KINEMATIC_UNIT_OBSERVER_H

#include <Core/core/system/ImportExport.h>
#include <Core/core/Component.h>
#include <Core/core/application/properties/Properties.h>
#include <Core/interface/observers/VariantBase.h>
#include <RobotAPI/interface/units/KinematicUnitInterface.h>
#include <RobotAPI/interface/observers/KinematicUnitObserverInterface.h>
#include <Core/observers/ConditionCheck.h>
#include <Core/observers/Observer.h>

#include <string>

namespace armarx
{
    ARMARX_CREATE_CHECK(KinematicUnitObserver, equals)
    ARMARX_CREATE_CHECK(KinematicUnitObserver, larger)

   /**
    * @class KinematicUnitObserver
    * @brief
    * @ingroup SensorActorUnits
    *
    * The KinematicUnitObserver allows to install conditions on all channel reported by the KinematicUnit.
    * These include joint angles, velocities, torques and motor temperatures
    *
    * The KinematicUnitObserver retrieves its configuration from a VirtualRobot robot model. Within the model, the joints
    * which are observer by the unit are define by a robotnodeset
    */
    class ARMARXCORE_IMPORT_EXPORT KinematicUnitObserver :
        virtual public Observer,
        virtual public KinematicUnitObserverInterface
    {
    public:
        // framework hooks
        void onInitObserver();
        void onConnectObserver();

        // slice interface implementation
        void reportControlModeChanged(const NameControlModeMap& jointModes, bool aValueChanged, const Ice::Current& c = ::Ice::Current());
        void reportJointAngles(const NameValueMap& jointAngles,  bool aValueChanged, const Ice::Current& c = ::Ice::Current());
        void reportJointVelocities(const NameValueMap& jointVelocities,  bool aValueChanged, const Ice::Current& c = ::Ice::Current());
        void reportJointTorques(const NameValueMap& jointTorques,  bool aValueChanged, const Ice::Current& c = ::Ice::Current());
        void reportJointCurrents(const NameValueMap& jointCurrents,  bool aValueChanged, const Ice::Current& c = ::Ice::Current());
        void reportJointMotorTemperatures(const NameValueMap& jointMotorTemperatures,  bool aValueChanged, const Ice::Current& c = ::Ice::Current());
        void reportJointStatuses(const NameStatusMap& jointStatuses, bool aValueChanged, const Ice::Current& c = ::Ice::Current());

        virtual std::string getDefaultName() const { return "KinematicUnitObserver"; };

        /**
         * @see PropertyUser::createPropertyDefinitions()
         */
        virtual PropertyDefinitionsPtr createPropertyDefinitions();

    protected:
        void nameValueMapToDataFields(const std::string& channelName, const NameValueMap& nameValueMap);

    private:
        std::string robotNodeSetName;
    };

    /**
      @class KinematicUnitDatafieldCreator
      @brief
      @ingroup SensorActorUnits
     */
    class KinematicUnitDatafieldCreator
    {
    public:
        KinematicUnitDatafieldCreator(const std::string &channelName): _channelName(channelName){}

        /**
         * @brief Function to create a Channel Representation for a KinematicUnitObserver
         * @param kinematicUnitOberserverName Name of the KinematicUnitObserver
         * @param jointName Name of a joint of the robot like it is specified
         * in the simox-robot-xml-file
         * @return returns a ChannelRepresentationPtr
         */
        DataFieldIdentifier getDatafield(const std::string &kinematicUnitObserverName, const std::string &jointName) const
        {
            if(kinematicUnitObserverName.empty())
                throw LocalException("kinematicUnitObserverName must not be empty!");
            if(jointName.empty())
                throw LocalException("jointName must not be empty!");
            return DataFieldIdentifier(kinematicUnitObserverName, _channelName, jointName);
        }

    private:
        std::string _channelName;
    };


    namespace channels{

        namespace KinematicUnitObserver{
            const KinematicUnitDatafieldCreator jointAngles("jointAngles");
            const KinematicUnitDatafieldCreator jointVelocities("jointVelocities");
            const KinematicUnitDatafieldCreator jointTorques("jointTorques");
            const KinematicUnitDatafieldCreator jointCurrents("jointCurrents");
            const KinematicUnitDatafieldCreator jointTemperatures("jointTemperatures");
        }
    }

}

#endif
