/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    ArmarXCore::units
 * @author     Christian Boege (boege dot at kit dot edu)
 * @date       2011
 * @copyright  http://www.gnu.org/licenses/gpl.txt
 *             GNU General Public License
 */

#ifndef _ARMARX_COMPONENT_KINEMATIC_UNIT_SIMULATION_H
#define _ARMARX_COMPONENT_KINEMATIC_UNIT_SIMULATION_H

#include "KinematicUnit.h"

#include <Core/core/services/tasks/PeriodicTask.h>
#include <Core/core/system/ImportExportComponent.h>

#include <IceUtil/Time.h>

namespace armarx
{
    /**
     * @class JointState.
     * @brief State of a joint.
     * @ingroup SensorActorUnits
     *
     * Includes the control mode, the joint angle, velocitym torquem and motor temperature.
     * Controlmode defaults to ePositionControl. All other values to zero.
     */
    class JointState
    {
    public:
        JointState()
        {
            init();
        }

        void init()
        {
            controlMode = ePositionControl;
            angle = 0.0f;
            velocity = 0.0f;
            torque = 0.0f;
            temperature = 0.0f;
        }

        ControlMode controlMode;
        float       angle;
        float       velocity;
        float       torque;
        float       temperature;
    };
    typedef std::map<std::string, JointState> JointStates;


    /**
     * @class JointInfo.
     * @brief
     * @ingroup SensorActorUnits
     *
     * Joint info including lower and upper joint limits (rad). No limits set by default (lo=hi).
     */
    class JointInfo
    {
    public:
        JointInfo()
        {
            reset();
        }

        void reset()
        {
            limitLo = 0.0f;
            limitHi = 0.0f;
        }
        bool hasLimits() const
        {
            return (limitLo != limitHi);
        }

        float limitLo;
        float limitHi;
    };
    typedef std::map<std::string, JointInfo> JointInfos;

    /**
     * @class KinematicUnitSimulationPropertyDefinitions
     * @brief
     * @ingroup SensorActorUnits
     */
    class KinematicUnitSimulationPropertyDefinitions :
            public KinematicUnitPropertyDefinitions
    {
    public:
        KinematicUnitSimulationPropertyDefinitions(std::string prefix) :
            KinematicUnitPropertyDefinitions(prefix)
        {
            defineOptionalProperty<int>("IntervalMs", 10, "The time in milliseconds between two calls to the simulation method.");
        }
    };

    /**
     * @class KinematicUnitSimulation
     * @brief
     * @ingroup SensorActorUnits
     */
    class KinematicUnitSimulation :
        virtual public KinematicUnit
    {
    public:
        // inherited from Component
        virtual std::string getDefaultName() const { return "KinematicUnitSimulation"; }

        virtual void onInitKinematicUnit();
        virtual void onStartKinematicUnit();
        virtual void onExitKinematicUnit();

        void simulationFunction();

        // proxy implementation
        virtual void requestJoints(const StringSequence& joints, const Ice::Current& c = ::Ice::Current());
        virtual void releaseJoints(const StringSequence& joints, const Ice::Current& c = ::Ice::Current());
        virtual void switchControlMode(const NameControlModeMap& targetJointModes, const Ice::Current& c = ::Ice::Current());
        virtual void setJointAngles(const NameValueMap& targetJointAngles, const Ice::Current& c = ::Ice::Current());
        virtual void setJointVelocities(const NameValueMap& targetJointVelocities, const Ice::Current& c = ::Ice::Current());
        virtual void setJointTorques(const NameValueMap& targetJointTorques, const Ice::Current& c = ::Ice::Current());
        virtual void setJointAccelerations(const NameValueMap& targetJointAccelerations, const Ice::Current& c = ::Ice::Current());
        virtual void setJointDecelerations(const NameValueMap& targetJointDecelerations, const Ice::Current& c = ::Ice::Current());
        void stop(const Ice::Current &c = Ice::Current());

        /**
         * @see PropertyUser::createPropertyDefinitions()
         */
        virtual PropertyDefinitionsPtr createPropertyDefinitions();

    protected:
        JointStates jointStates;
        JointStates previousJointStates;
        JointInfos jointInfos;
        boost::mutex jointStatesMutex;
        IceUtil::Time lastExecutionTime;
        int intervalMs;

        PeriodicTask<KinematicUnitSimulation>::pointer_type simulationTask;
    };
}

#endif
