/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    ArmarXCore::units
 * @author     Christian Boege (boege at kit dot edu)
 * @date       2011
 * @copyright  http://www.gnu.org/licenses/gpl.txt
 *             GNU General Public License
 */


#include "TCPMoverUnit.h"

#include <RobotAPI/robotstate/remote/ArmarPose.h>
#include <Core/core/system/ArmarXDataPath.h>

//VirtualRobot
#include <VirtualRobot/XML/RobotIO.h>
#include <VirtualRobot/LinkedCoordinate.h>




/** VirtualRobot headers **/
#include <VirtualRobot/Robot.h>
#include <VirtualRobot/Nodes/RobotNode.h>
#include <VirtualRobot/RobotNodeSet.h>
#include <VirtualRobot/IK/DifferentialIK.h>

#include <Eigen/Geometry>

using namespace armarx;
using namespace std;
using namespace VirtualRobot;


TCPMoverUnit::TCPMoverUnit():
      selectedHand(eNone)

{
    handData[eLeftHand];
    handData[eRightHand];
}


void TCPMoverUnit::onInitComponent()
{
    kinematicUnitFile = getProperty<std::string>("RobotFileName").getValue();
    kinematicUnitName = getProperty<std::string>("RobotNodeSetName").getValue();
    if (!ArmarXDataPath::getAbsolutePath(kinematicUnitFile,kinematicUnitFile))
    {
        throw UserException("Could not find robot file " + kinematicUnitFile);
    }
    // Load Robot
    ARMARX_VERBOSE << "Loading KinematicUnit " << kinematicUnitName << " from '" << kinematicUnitFile << "' ..." << flush;


    robot = RobotIO::loadRobot(kinematicUnitFile);
    if (!robot)
    {
        ARMARX_ERROR << "Could not find Robot XML file with name '" << kinematicUnitFile << "'" << flush;
    }


    usingProxy(kinematicUnitName);

    periodicTask = new PeriodicTask<TCPMoverUnit>(this, &TCPMoverUnit::periodicExec, 40, false, "TCPMoverUnit");
}


void TCPMoverUnit::onConnectComponent()
{
    kinematicUnitPrx = getProxy<KinematicUnitInterfacePrx>(kinematicUnitName);


    handData[eLeftHand].setTCPNodeSet(robot->getRobotNodeSet("LARM"));
    handData[eRightHand].setTCPNodeSet(robot->getRobotNodeSet("RARM"));
    resetToHomePosition(eLeftHand);
    resetToHomePosition(eRightHand);

    ARMARX_VERBOSE << "loaded";
    periodicTask->start();
}


void TCPMoverUnit::onExitComponent()
{
    try{
        if(robotRequested && kinematicUnitPrx)
            kinematicUnitPrx->release();
    }catch(Ice::NotRegisteredException e)
    {
//        ARMARX_ERROR << "Lost connection - shutting down" << flush;
//        this->interrupt();
    }
    periodicTask->stop();
}


void TCPMoverUnit::periodicExec()
{
    this->calculateNewHandPosition(eLeftHand);
    this->calculateNewHandPosition(eRightHand);
    return;


}

void TCPMoverUnit::request(const Ice::Current &)
{
    try{
        kinematicUnitPrx->request();
        robotRequested = true;
    }catch(ResourceUnavailableException &e)
    {
        ARMARX_WARNING << "Resource is unavailable" << flush;
    }
}

void TCPMoverUnit::release(const Ice::Current &)
{
    try
    {
        kinematicUnitPrx->release();
        robotRequested = false;
    }
    catch(ResourceNotOwnedException &e)
    {
        ARMARX_WARNING << "Resource is not owned" << flush;
    }
}


void TCPMoverUnit::resetArmToHomePosition(HandSelection selectedHand, const Ice::Current &)
{
    resetToHomePosition(selectedHand);
}


void TCPMoverUnit::setCartesianTCPVelocity(HandSelection selectedHand, Ice::Float x, Ice::Float y, Ice::Float z, Ice::Float speedFactor, const Ice::Current &)
{
    if(selectedHand != eLeftHand && selectedHand != eRightHand)
        throw LocalException("Please select a specific hand.");
    if(handData.find(selectedHand) == handData.end())
        throw LocalException("HandData not found.");
    HandData  &hand  = handData[selectedHand];
    ARMARX_VERBOSE << "setting new cartesian tcp velocity";
    this->selectedHand = selectedHand;
    hand.velocities[0] = x*speedFactor;
    hand.velocities[1] = y*speedFactor;
    hand.velocities[2] = z*speedFactor;

}


void TCPMoverUnit::setTCPOrientation(HandSelection selectedHand, Ice::Float alpha, Ice::Float beta, Ice::Float gamma, const Ice::Current &)
{
    if(selectedHand != eLeftHand && selectedHand != eRightHand)
        throw LocalException("Please select a specific hand.");
    if(handData.find(selectedHand) == handData.end())
        throw LocalException("HandData not found.");
    HandData  &hand  = handData[selectedHand];
    ARMARX_VERBOSE << "setting new tcp orientation";
    this->selectedHand = selectedHand;
    hand.orientation[0] = alpha;
    hand.orientation[1] = beta;
    hand.orientation[2] = gamma;
}


void TCPMoverUnit::setPlatformVelocities(Ice::Float x, Ice::Float y, Ice::Float theta, const Ice::Current &)
{
}

void TCPMoverUnit::setRotationMat(Eigen::Matrix4f &matrix, float alpha, float beta, float gamma)
{
     using namespace Eigen;
    const float sinfalpha = sin(alpha);
    const float cosfalpha = cos(alpha);
    const float sinfbeta = sin(beta);
    const float cosfbeta = cos(beta);
    const float sinfgamma = sin(gamma);
    const float cosfgamma = cos(gamma);

    matrix(0,0) = cosfbeta * cosfgamma;
    matrix(0,1) = - cosfbeta * sinfgamma;
    matrix(0,2) = sinfbeta;
    matrix(1,0) = cosfalpha * sinfgamma + sinfalpha * sinfbeta * cosfgamma;
    matrix(1,1) = cosfalpha * cosfgamma - sinfalpha * sinfbeta * sinfgamma;
    matrix(1,2) = - sinfalpha * cosfbeta;
    matrix(2,0) = sinfalpha * sinfgamma - cosfalpha * sinfbeta * cosfgamma;
    matrix(2,1) = sinfalpha * cosfgamma + cosfalpha * sinfbeta * sinfgamma;
    matrix(2,2) = cosfalpha * cosfbeta;
}


void TCPMoverUnit::resetToHomePosition(HandSelection selectedHand)
{
    if(selectedHand != eLeftHand && selectedHand != eRightHand)
        throw LocalException("Please select a specific hand.");
    if(handData.find(selectedHand) == handData.end())
        throw LocalException("HandData not found.");
    HandData  &hand  = handData[selectedHand];
    hand.orientation.resize(3, 0.f);
    std::vector<float> targetVec(19,0.f);
    targetVec[11] = -0.009199;
    targetVec[14] = -3.14/2;
    targetVec[15] = -90.f/180*3.145;
    if(selectedHand == eRightHand)
        targetVec[15] *= -1;

    robot->setJointValues(hand.getTCPNodeSet(),targetVec);
    hand.standardRot = hand.getTCPNodeSet()->getTCP()->getGlobalPoseVisualization();
}

void TCPMoverUnit::setHandConfiguration(HandSelection, Ice::Int, const Ice::Current &)
{
    ARMARX_INFO << "not yet implemented" << std::endl;
}


void TCPMoverUnit::calculateNewHandPosition(HandSelection selectedHand)
{
    if(selectedHand != eLeftHand && selectedHand != eRightHand)
        throw LocalException("Please select a specific hand.");
    if(handData.find(selectedHand) == handData.end())
        throw LocalException("HandData not found.");
    HandData  &hand  = handData[selectedHand];
    RobotNodePtr tcpNode = hand.getTCPNodeSet()->getTCP();
    using namespace Eigen;


    Matrix4f pose = tcpNode->getGlobalPoseVisualization();
//    cout << "angles: \n" << orientation[0] << ", " << orientation[1] << ", " << orientation[2] << endl;
    Matrix4f targetPose;
    Matrix4f xRot;
    xRot << 1, 0, 0, 0,
            0, cos(hand.orientation[0]), -sin(hand.orientation[0]), 0,
            0, sin(hand.orientation[0]), cos(hand.orientation[0]), 0,
            0, 0, 0, 1;

    Matrix4f yRot;
    yRot << cos(hand.orientation[1]), 0, sin(hand.orientation[1]), 0,
            0, 1, 0, 0,
            -sin(hand.orientation[1]), 0, cos(hand.orientation[1]), 0,
            0, 0, 0, 1;


    Matrix4f zRot;
    zRot <<
            cos(hand.orientation[2]), -sin(hand.orientation[2]), 0, 0,
            sin(hand.orientation[2]), cos(hand.orientation[2]), 0, 0,
            0, 0, 1, 0,
            0, 0, 0, 1;


    targetPose = xRot*yRot*zRot*hand.standardRot;
    targetPose(0,3) = pose(0,3) + hand.velocities[0];
    targetPose(1,3) = pose(1,3) + hand.velocities[1];
    targetPose(2,3) = pose(2,3) + hand.velocities[2];
//    std::cout << targetPose << std::endl;
    hand.ik->setGoal(targetPose, tcpNode, IKSolver::All);

    VectorXf errorJoint(hand.getTCPNodeSet()->getSize());
    errorJoint = hand.ik->computeStep(0.1f);

    // assign new values to local robot and to kinematicUnit
    std::vector<float> angles = hand.getTCPNodeSet()->getJointValues();
    std::vector<float> targetAngles(hand.getTCPNodeSet()->getSize());
    NameValueMap targetMap;

    for(unsigned int i = 0; i < hand.getTCPNodeSet()->getSize(); i++)
    {
        float newAngle = angles[i] + errorJoint(i);
        targetAngles[i] = newAngle;
        targetMap[hand.getTCPNodeSet()->getNode(i)->getName()] = newAngle;
        //ARMARX_VERBOSE << tcpNodeSet->getNode(i)->getName() << ": " << newAngle;
    }
    robot->setJointValues(hand.getTCPNodeSet(), targetAngles);

    try{
        kinematicUnitPrx->setJointAngles(targetMap);
    }catch(Ice::NotRegisteredException e)
    {
        ARMARX_ERROR << "Lost connection - shutting down" << flush;
        this->terminate();
    }
}


void TCPMoverUnit::lookWithHeadTo(Ice::Float y, Ice::Float, Ice::Float, const Ice::Current &)
{
}


void TCPMoverUnit::HandData::setTCPNodeSet(RobotNodeSetPtr tcpNodeSet)
{
    this->tcpNodeSet = tcpNodeSet; ik =  VirtualRobot::DifferentialIKPtr(new DifferentialIK(tcpNodeSet));
}


PropertyDefinitionsPtr TCPMoverUnit::createPropertyDefinitions()
{
    return PropertyDefinitionsPtr(new TCPMoverUnitPropertyDefinitions(
                                           getConfigIdentifier()));
}
